﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using RGBRepair;

namespace RgbRepair.ColorProcessing {
	public static class ColorProcessingManager {
		private readonly static float moreThanThreshold = 0.5f;
		private readonly static float lessThanThreshold = 0.5f;

		private readonly static ColorCriteriaEntry[] colorCriteriaLibrary = new ColorCriteriaEntry[] {
			new ColorCriteriaEntry (
					EColor.Red,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.LessThan, lessThanThreshold)
					}
				),
			new ColorCriteriaEntry (
					EColor.Blue,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.MoreThan, moreThanThreshold)
					}
				),
			new ColorCriteriaEntry (
					EColor.Green,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.LessThan, lessThanThreshold)
					}
				),
			new ColorCriteriaEntry (
					EColor.Yellow,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.LessThan, lessThanThreshold)
					}
				),
			new ColorCriteriaEntry (
					EColor.Purple,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.MoreThan, moreThanThreshold)
					}
				),
			new ColorCriteriaEntry (
					EColor.White,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.MoreThan, moreThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.MoreThan, moreThanThreshold)
					}
				),
			new ColorCriteriaEntry (
					EColor.Black,
					new ColorCriterion[] {
						new ColorCriterion (EColorPart.Red, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Green, EColorComparator.LessThan, lessThanThreshold),
						new ColorCriterion (EColorPart.Blue, EColorComparator.LessThan, lessThanThreshold)
					}
				)
		};

		/// <summary>
		/// Determines a main color out of the center pixels of a Texture2D
		/// </summary>
		/// <param name="texture">The texture to process</param>
		/// <param name="crosshairRadius">Determines how big the sample taken from the center will be</param>
		/// <param name="maxErrorTolerance">How many, in percentage of the total, pixels can be ignored when determining main color. Preferably not below 50%</param>
		/// <returns>The main color of the sampled pixels</returns>
		public static EColor GetMainColorOfTexture (Texture2D texture, int crosshairRadius, float maxErrorTolerance) {
			Color[] centerPixels = GetCenterPixels (texture, crosshairRadius);

			foreach (ColorCriteriaEntry criteriaEntry in colorCriteriaLibrary) {
				bool checkResult = DoPixelsFitCriteria (centerPixels, criteriaEntry.criteria, maxErrorTolerance);
				if (checkResult) {
					return criteriaEntry.colorName;
				}
			}

			return EColor.Undefined;
		}

		private static Color[] GetCenterPixels (Texture2D texture, int radius) {
			int centerX = texture.width / 2;
			int centerY = texture.height / 2;
			int resultSize = (radius * 2 + 2) * (radius * 2 + 2);
			List<Color> resultPixels = new List<Color> ();
			for (int i = centerX - radius; i <= centerX + radius; i++) {
				for (int j = centerY - radius; j <= centerY + radius; j++) {
					resultPixels.Add (texture.GetPixel (i, j));
				}
			}
			return resultPixels.ToArray ();
		}

		public static bool DoPixelsFitCriteria (
			Color[] pixels,
			ColorCriterion[] criteria,
			float maxErrorPercentage) {

			int trueCount = 0, falseCount = 0;
			foreach (Color pixel in pixels) {
				if (pixel == null) continue;

				bool doPixelFitAllCriterion = true;

				foreach (ColorCriterion criterion in criteria) {
					if (!DoPixelFitCriterion(pixel, criterion)) {
						doPixelFitAllCriterion = false;
					}
				}

				if (doPixelFitAllCriterion) {
					trueCount++;
				} else {
					falseCount++;
				}
			}

			if (trueCount + falseCount == 0) return false;

			float errorPercentage = falseCount * 1.0f / (trueCount + falseCount) * 1.0f;
			return errorPercentage < maxErrorPercentage;
		}

		private static bool DoPixelFitCriterion (Color pixel, ColorCriterion criterion) {
			float colorValue = 0f;
			switch (criterion.colorPart) {
				case EColorPart.Red: {
						colorValue = pixel.r;
					}
					break;
				case EColorPart.Blue: {
						colorValue = pixel.b;
					}
					break;
				case EColorPart.Green: {
						colorValue = pixel.g;
					}
					break;
			}
			switch (criterion.comparator) {
				case EColorComparator.LessThan: {
						return colorValue < criterion.colorThreshold;
					} break;
				case EColorComparator.MoreThan: {
						return colorValue > criterion.colorThreshold;
					} break;
			}
			return false;
		}

		public static EColor ConvertMaterialToEColor (iMaterial material) {
			if (material.r > 0.5f) {
				if (material.g > 0.5f) {
					if (material.b > 0.5f) {
						return EColor.White;
					} else {
						return EColor.Yellow;
					}
				} else {
					if (material.b > 0.5f) {
						return EColor.Purple;
					} else {
						return EColor.Red;
					}
				}
			} else {
				if (material.g > 0.5f) {
					if (material.b > 0.5f) {
						//cyan not implemented
						return EColor.Undefined;
					} else {
						return EColor.Green;
					}
				} else {
					if (material.b > 0.5f) {
						return EColor.Blue;
					} else {
						return EColor.Black;
					}
				}
			}
		}
	}
}
