﻿using System;
using System.IO;
using UnityEngine;

namespace Seven
{
    public static class IOExtend
    {
        public static void WriteToFile(this byte[] bytes, string path)
        {
            if (bytes.IsNull() || path.IsNullOrEmpty())
                return;

            try
            {
                string dir = Path.GetDirectoryName(path);

                if (!Directory.Exists(dir))
                    Directory.CreateDirectory(dir);

                File.WriteAllBytes(path, bytes);
            }
            catch (Exception ex)
            {
                ex.Message.Log();
            }
        }
        public static void WriteToFile(this string str, string path)
        {
            if (path.IsNullOrEmpty())
                return;

            try
            {
                string dir = Path.GetDirectoryName(path);

                if (!Directory.Exists(dir))
                    Directory.CreateDirectory(dir);

                File.WriteAllText(path, str);
            }
            catch (Exception ex)
            {
                ex.Message.Log();
            }
        }
        public static void WriteToFile(this Texture2D texture, string path)
        {
            if (path.IsNullOrEmpty())
                return;

            string extension = Path.GetExtension(path).ToLowerInvariant();

            switch (extension)
            {
                case ".jpg":
                    {
                        texture.EncodeToJPG().WriteToFile(path);
                        break;
                    }

                case ".jpeg":
                    {
                        texture.EncodeToJPG().WriteToFile(path);
                        break;
                    }

                case ".png":
                    {
                        texture.EncodeToPNG().WriteToFile(path);
                        break;
                    }

                case ".tga":
                    {
                        texture.EncodeToTGA().WriteToFile(path);
                        break;
                    }

                case ".exr":
                    {
                        texture.EncodeToEXR().WriteToFile(path);
                        break;
                    }

                default:
                    {
                        texture.EncodeToPNG().WriteToFile(path);
                        break;
                    }
            }
        }
        public static void OpenFileWithPath(this string path)
        {
            if (!File.Exists(path))
                return;

            try
            {
                System.Diagnostics.Process.Start(Path.GetFullPath(path));
            }
            catch (Exception ex)
            {
                ex.Message.LogError();
            }
        }
        public static void OpenFolderWithPath(this string path)
        {
            try
            {
                string directory = Path.GetDirectoryName(path);

                if (Directory.Exists(directory))
                {
                    System.Diagnostics.Process.Start(directory);
                }
            }
            catch (Exception ex)
            {
                ex.Message.LogError();
            }
        }
        public static string ReadFile(this string path)
        {
            if (path.IsNullOrEmpty() || !File.Exists(path))
                return null;

            return File.ReadAllText(path);
        }
        public static byte[] ReadFileBytes(this string path)
        {
            if (path.IsNullOrEmpty() || !File.Exists(path))
                return null;

            return File.ReadAllBytes(path);
        }
        public static bool FileExist(this string path)
        {
            if (path.IsNullOrEmpty())
                return false;

            return File.Exists(path);
        }
        public static void RemoveFile(this string path)
        {
            if (path.IsNullOrEmpty() || !File.Exists(path))
                return;

            File.Delete(path);
        }
        public static string CombinePath(this string path, string pathB)
        {
            return Path.Combine(path, pathB);
        }
        public static void CopyFileTo(this string path, string destination)
        {
            if (path.IsNullOrEmpty() || destination.IsNullOrEmpty())
                return;

            File.Copy(path, destination, true);
        }
        public static void CreateDirectoryIfNotExist(this string path)
        {
            if (path.IsNullOrEmpty())
                return;

            string dir = Path.GetDirectoryName(path);
            if (!Directory.Exists(dir))
            {
                Directory.CreateDirectory(dir);
            }
        }
        public static void RemoveDirectory(this string path)
        {
            if (path.IsNullOrEmpty() || !Directory.Exists(path))
                return;

            Directory.Delete(path);
        }
        public static string ReplaceDirectorySeparator(this string str, char replace = '/')
        {
            if (str.Contains("\\"))
            {
                return str.Replace('\\', replace);
            }
            else
            {
                return str;
            }
        }
    }
}
